# _METADATA_:Version: 20
# _METADATA_:Timestamp: 2021-01-17 21:26:27.469550+00:00
# _METADATA_:MD5: 1d26a56d8a45619d63b5715fb215c706
# _METADATA_:Publish:                                                                      None


# _METADATA_:
import logging
from datetime import datetime, timedelta, timezone
from neo4j import GraphDatabase, basic_auth, __version__ as neoVersion
from neo4j.exceptions import ServiceUnavailable
import uuid
import redis

from serpentmonkee.PubSubMonkee import PubSubMonkee
from serpentmonkee.CypherQueue import CypherQueue, CypherQueues
from serpentmonkee.CypherTransaction import CypherTransactionBlock, CypherTransactionBlockWorker
from serpentmonkee.MonkeeSQLblocks import MonkeeSQLblock, MonkeeSQLblockHandler


class NeoMonkee:  # --------------------------------------------------------------------

    def __init__(self, neoDriver, redisClient, publisher, projectId, topicId, sqlTable, callingCF=None):
        self.neoDriver = neoDriver
        self.driverUuid = None
        self.driverStartedAt = None
        self.sqlTable = sqlTable
        self.callingCF = callingCF
        self.redisClient = redisClient
        self.db_fb = None
        self.cypherQueues = self.makeCypherQueues()
        self.asyncStatements = []
        self.pubsub = PubSubMonkee(publisher, projectId, topicId)
        self.cypherWorker = CypherTransactionBlockWorker(
            self.neoDriver, self.cypherQueues, pubsub=self.pubsub, redisClient=redisClient, environmentName=projectId)

        self.sqlBlockHandler = MonkeeSQLblockHandler(environmentName=projectId,
                                                     redis_client=redisClient,
                                                     pubsub=publisher)
        # self.sqlBlockHandler.killQueue()

        # self.sqlWorker = MonkeeSQLblockWorker(environmentName=projectId,
        #                                      sqlClient=sqlClient,
        #                                      sqlBHandler=self.sqlBlockHandler)

        """
        sqlb = MonkeeSQLblock(query=sqlInsertQuery,
                                  insertList=[],
                                  queryTypeId='humans')
        self.sqlBlockHandler.toQ(sqlb)
        """

    def makeCypherQueues(self):
        cQH = CypherQueue("cypherQ_High")
        cQM = CypherQueue("cypherQ_Medium")
        cQL = CypherQueue("cypherQ_Low")
        wQ = CypherQueue("cypherWorking")
        compQ = CypherQueue("cypherDone")
        queues = [cQH, cQM, cQL]
        return CypherQueues(redisClient=self.redisClient,
                            cQueues=queues, workingQ=wQ, completedQ=compQ, fb_db=self.db_fb)

    def get_uuid(self):
        return str(uuid.uuid4())

    def makeNeoDriver(self, neo_uri, neo_user, neo_pass):
        if neo_uri is not None:
            self.driverUuid = self.get_uuid()
            self.driverStartedAt = datetime.now(timezone.utc)
            if neoVersion[0] == '4':
                self.neoDriver = GraphDatabase.driver(
                    uri=neo_uri,
                    auth=basic_auth(
                        user=neo_user,
                        password=neo_pass,
                    ),
                    max_transaction_retry_time=2
                    # max_connection_lifetime=200,
                    # encrypted=True,
                )
            if neoVersion[0] == '1':
                self.neoDriver = GraphDatabase.driver(
                    uri=neo_uri,
                    auth=basic_auth(
                        user=neo_user,
                        password=neo_pass,
                    ),
                    # max_connection_lifetime=200,
                    encrypted=True,
                    max_retry_time=2)

    def syncRead(self, query, cfInstanceUid='', **params):
        """
        Reads the results of a cypher query.

            USAGE:

            query = '''
                    MATCH (h:nodeType {_project: $projectname })
                    return h.uid as uid
                    '''

            params = {'projectname': 'Sandbox'}
            res = neomnkee.syncRead(query=query, params=params)
            for r in res:
                print(r['uid'])
        """
        start_ts = datetime.now(timezone.utc)
        with self.neoDriver.session() as session:
            result = session.read_transaction(
                self._inner,
                query,
                cfInstanceUid,
                **params,
            )

        end_ts = datetime.now(timezone.utc)
        self.saveToSql(proc_name='syncRead',
                       start_ts=start_ts,
                       end_ts=end_ts,
                       cypher=query,
                       params=str(params),
                       batch=str([]),
                       cfInstanceUid=cfInstanceUid)
        return result

    def asyncWriteStatement(self, query, batch=[], cfInstanceUid='', **params):

        if batch == []:
            self.asyncStatements.append(
                {"cypher": query, "parameters": params['params'], "ts": datetime.now(timezone.utc)})
        else:
            self.asyncStatements.append(
                {"cypher": query, "parameters": params['params'], "batch": batch, "ts": datetime.now(timezone.utc)})

    def asyncWrite(self, priority="M", appUid=None, docUid=None):
        """
        batches all the async statements posted so far and sends it off to the cypherQueue identified by priority
        """
        if self.asyncStatements == []:
            print('XXX no statements to action')
        else:
            guid = self.get_uuid()

            ctb = CypherTransactionBlock(priority=priority, statements=self.asyncStatements,
                                         transactionUid=guid, callingCF=self.callingCF, originDocUid=docUid, appUid=appUid, sqlBlockHandler=self.sqlBlockHandler)
            self.cypherQueues.pushCtbToWaitingQ(ctb)
            # Resets the asyncStatements: this "asyncWrite" is seen as a push the statements to queue and start with a clean list
            self.asyncStatements = []
            self.cypherQueues.getQLens()

            # Sends a pubsub message to start the cypher worker CF
            self.pubsub.publish_message('awaken')

    def syncWrite(self, query, cfInstanceUid='', **params):
        start_ts = datetime.now(timezone.utc)
        result = None
        with self.neoDriver.session() as session:
            result = session.write_transaction(self._inner, query,
                                               cfInstanceUid, **params)

        end_ts = datetime.now(timezone.utc)
        self.saveToSql(proc_name='syncWrite',
                       start_ts=start_ts,
                       end_ts=end_ts,
                       cypher=query,
                       params=str(params),
                       batch=str([]),
                       cfInstanceUid=cfInstanceUid)
        return result

    def _inner(self, tx, query, cfInstanceUid, params):
        result = tx.run(query, params)

        try:
            return [row for row in result]
        except ServiceUnavailable as e:
            self.saveToSql('_inner', None, None, repr(e),
                           'ERROR: ServiceUnavailable', None, cfInstanceUid)
            logging.error(repr(e))
            raise
        except Exception as e:
            self.saveToSql('_inner', None, None, repr(e), 'ERROR: Other', None,
                           cfInstanceUid)
            logging.error(repr(e))
            raise

    def syncWriteBatch(self, query, batch, cfInstanceUid='', **params):
        """
        Runs a batch update.

            USAGE:
            batch = ['01602cb23de544e8b33c4612810e96a5',
                '016a8a2f414c43b49b656b655da07fbe',
                '01f62dc44f6d4e85b0c2a7a973f97750']
            query = '''
                    UNWIND $batch AS hB
                    MATCH (h:humans { uid: hB, _project: $projectname })
                    set h.val = 314159
                    return h.uid as uid
                    '''

            params = {'projectname': 'Sandbox'}
            res = neomnkee.syncWriteBatch(
                query=query, params=params, batch=batch)
            for r in res:
                print(r['uid'])
        """
        write_results = None
        start_ts = datetime.now(timezone.utc)
        with self.neoDriver.session() as session:
            write_results = session.write_transaction(
                self._innerBatch,
                query,
                batch,
                cfInstanceUid,
                **params,
            )

        end_ts = datetime.now(timezone.utc)
        self.saveToSql(proc_name='syncWriteBatch',
                       start_ts=start_ts,
                       end_ts=end_ts,
                       cypher=query,
                       params=str(params),
                       batch=str(batch),
                       cfInstanceUid=cfInstanceUid)
        return write_results

    def _innerBatch(self, tx, query, batch, cfInstanceUid, params):
        result = tx.run(
            query,
            params,
            batch=batch,
        )

        try:
            return [row for row in result]
        except ServiceUnavailable as e:
            logging.error(repr(e))
            self.saveToSql('_innerBatch', None, None, repr(e),
                           'ERROR: ServiceUnavailable', None, cfInstanceUid)
            raise
        except Exception as e:
            self.saveToSql('_inner', None, None, repr(e), 'ERROR: Other', None,
                           cfInstanceUid)
            logging.error(repr(e))
            raise

    def saveToSql(self, proc_name, start_ts, end_ts, cypher, params, batch,
                  cfInstanceUid):
        """inserts the payloads to the necessary SQL tables"""
        try:

            timeFormat = "%Y-%m-%d, %H:%M:%S"
            duration = None

            if start_ts is not None and end_ts is not None:
                timeDiff = end_ts - start_ts
                duration = timeDiff.total_seconds()
            else:
                start_ts = datetime.now(timezone.utc)
                end_ts = datetime.now(timezone.utc)

            sqlInsertQuery = """ INSERT INTO """ + self.sqlTable + """(procedure_name,query_start,query_end,query_duration_in_s,cypher,params,batch,connector_uid,connector_start_time,calling_cf,cf_instance_uid)
            VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
            """
            sqlInsertList = [
                proc_name,
                start_ts.strftime(timeFormat),
                end_ts.strftime(timeFormat), duration, cypher,
                params,
                str(batch), self.driverUuid, self.driverStartedAt,
                self.callingCF, cfInstanceUid
            ]

            sqlb = MonkeeSQLblock(query=sqlInsertQuery,
                                  insertList=sqlInsertList,
                                  queryTypeId='')
            self.sqlBlockHandler.toQ(sqlb, priority='H')

        except Exception as e:
            logging.error("saveToSql: {}".format(repr(e)))
            raise



# _METADATA_:Version: 15
# _METADATA_:Timestamp: 2021-06-18 16:11:18.297375+00:00
# _METADATA_:MD5: f05813b3e6cbf0196dbacc4eec8fbdb1
# _METADATA_:Publish:                                                                                None


# _METADATA_:
from datetime import datetime, timedelta, timezone
import time
import logging
import re


class MonkeeRedis:
    """
    Wrapper for dealing with redis.
    - fb_db:  Firebase client. if None, session_id checks cannot be done
    - cfName: the name of the calling CF. If None, checks cannot be done for the last CF call
    """

    def __init__(
        self, fb_db, cfName, redisClient, appUid='', userUid='', inDebugMode=False
    ):
        self.fb_db = fb_db
        self.callingCF = cfName
        self.redis = redisClient
        self.userUid = userUid
        self.appUid = appUid
        self.inDebugMode = inDebugMode
        self.sessionId = None
        self.timestamp = datetime.now(timezone.utc)
        self.minsBetweenSessions = 30
        self.timeSinceLastEvent = None

    def dprint(self, stringg):
        if self.inDebugMode:
            print(stringg)

    def set_play_event(self, contentUid, playTime):
        """
        Sets the time that the content was played/started
        """
        self.dprint(
            "setting set_play_event = {} - {}".format(contentUid, playTime))
        fieldName = "ContentPlay={}".format(contentUid)
        self.set_project_human_val(
            fieldName=fieldName, dataType="datetimeNTZ", value=playTime, expireInSeconds=60 * 60
        )

    def kill_play_event(self, contentUid):
        fieldName = "ContentPlay={}".format(contentUid)
        self.kill_project_human_val(fieldName=fieldName)

    def get_play_event(self, contentUid):
        """
        Gets the last time playEvent for this content for this user, assuming that it's still in Redis
        """
        self.dprint("getting set_play_event = {} ".format(contentUid))
        fieldName = "ContentPlay={}".format(contentUid)
        return self.get_project_human_val(fieldName=fieldName)

    def get_time_in_content(self, lastPlayTime, endTime):
        """
        Calcs the time-in-seconds spent in the content
        """

        if lastPlayTime is None:
            self.dprint(
                "get_time_in_content ={}.".format("None"))
            return None
        else:
            timeDiff = endTime - lastPlayTime
            self.dprint("get_time_in_content ={}".format(
                timeDiff.total_seconds()))
            return timeDiff.total_seconds()

    def set_last_cf_call(self, eventTime):
        """Sets the last time that this CF was called in this project by this particular user"""
        if self.callingCF:
            self.dprint("setting last_cf_call = {}".format(self.callingCF))
            fieldName = "CFcall={}".format(self.callingCF)
            self.set_project_human_val(
                fieldName=fieldName, dataType="datetimeNTZ", value=eventTime, expireInSeconds=60 * 60 * 24 * 30
            )
        else:
            self.dprint("set_last_cf_call: no CF name specified")

    def get_last_cf_call(self):
        """Gets the last time that this CF was called in this project by this particular user"""
        if self.callingCF:
            fieldName = "CFcall={}".format(self.callingCF)
            return self.get_project_human_val(fieldName=fieldName)
        return None

    def get_sec_since_last_cf_call(self, eventTime):
        llt = self.get_last_cf_call()

        if llt is None:
            self.dprint(
                "get_sec_since_last_cf_call ={}. Setting it now".format("None"))
            self.set_project_human_val(
                fieldName="lastLogTime", dataType="datetimeNTZ", value=eventTime
            )

            return 100
        else:
            if not eventTime.tzinfo and not llt.tzinfo:
                timeDiff = eventTime - llt
            else:
                timeDiff = self.timestamp - llt
            self.dprint("get_sec_since_last_cf_call ={}".format(
                timeDiff.total_seconds()))
            return timeDiff.total_seconds()

    def kill_project_human_val(self, fieldName):
        """
        Removes the given key from Redis
        """
        key = self.appUid + ":" + self.userUid + ":" + fieldName
        if self.redis is not None:
            self.redis.delete(key)

    def set_project_human_val(self, fieldName, dataType, value, expireInSeconds=None):
        """Sets the compound key [self.appUid + ":" + self.userUid + ":" + fieldName] to
        value [dataType + "|" + str(value)]

        - dataType one of ["datetime", "datetimeNTZ", "int", "str"]. 
        """
        key = self.appUid + ":" + self.userUid + ":" + fieldName
        val = dataType + "|" + str(value)
        self.dprint("set_project_human_val: key={}, val={}".format(key, val))
        if self.redis is not None:
            self.redis.set(key, val)
            if expireInSeconds is not None:
                self.redis.expire(key, expireInSeconds)

    def get_project_human_val(self, fieldName):
        """Gets the compound key [self.appUid + ":" + self.userUid + ":" + fieldName]
        from redis and casts it back to native based on the [dataType] used in its initial storing
        """
        if self.redis is not None:
            key = self.appUid + ":" + self.userUid + ":" + fieldName
            val = self.redis.get(key)
            if val is None or val == "None":
                return None
            else:
                val = val.decode("utf-8")
            splitted = val.split("|")
            self.dprint("get_project_human_val splitted: {} ".format(splitted))
            dataType = splitted[0]
            value = splitted[1]
            self.dprint(
                "get_project_human_val: {} = {} ({})".format(
                    key, value, dataType)
            )
            # if value[-6]='+' and value[-3]==":" and dataType='datetime':

            return self.format_val(dataType, value)
        else:
            return None

    def format_val(self, dataType, val):
        """Formats the [val] value given the [dataType]"""
        if val is None:
            return None
        try:
            if dataType in ["datetime", "datetimeNTZ"]:
                fmt = self.inferDTFormat(val)
                return datetime.strptime(val, fmt)
            elif dataType == "int":
                return int(val)
            else:
                return val
        except ValueError as e:
            logging.error(repr(e))
            return None

    def get_session_id(self):
        self.sessionId = self.get_project_human_val("sessionId")
        self.dprint("get_session_id={}".format(self.sessionId))
        return self.sessionId

    def calc_session_id(self, time_diff):
        self.get_session_id()
        if self.sessionId is None:
            self.sessionId = 1
            self.set_session_id()
            return self.sessionId, None, None
        elif time_diff > self.minsBetweenSessions * 60:
            self.sessionId += 1
            self.dprint("incrementing session ID to {}".format(self.sessionId))
            self.set_session_id()
            self.timeSinceLastEvent = time_diff
            return self.sessionId, self.timeSinceLastEvent, True
        elif time_diff <= self.minsBetweenSessions * 60:
            self.dprint("keeping session ID as {}".format(self.sessionId))
            self.timeSinceLastEvent = time_diff
            return self.sessionId, self.timeSinceLastEvent, False

    def set_session_id(self):
        self.set_project_human_val("sessionId", "int", self.sessionId)

        if self.fb_db and self.sessionId >= 1 and len(self.userUid) > 2:
            humanResultRef = self.fb_db.collection(
                "apps", self.appUid, "humans", self.userUid, "results"
            ).document("session")

            humanResultRef.set(
                {
                    "sessionNumber": self.sessionId,
                    "src": "MonkeeRedis from monkee_log_v2",
                    "session": {
                        "tag": "session" + str(self.sessionId),
                        "answeredAt": self.timestamp,
                        "src": "MonkeeRedis from monkee_log_v2"
                    },
                },
                merge=True,
            )

    def inferDTFormat(self, dtString):
        dtFormats = []
        dtFormats.append({"re": re.compile(
            r"\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\.\d*\+\d{2}:\d{2}$"), "fmt": "%Y-%m-%d %H:%M:%S.%f%z"})
        dtFormats.append({"re": re.compile(
            r"\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\.\d*$"), "fmt": "%Y-%m-%d %H:%M:%S.%f"})
        dtFormats.append({"re": re.compile(
            r"\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}$"), "fmt": "%Y-%m-%d %H:%M:%S"})

        for f in dtFormats:
            if len(re.findall(f["re"], dtString)) > 0:
                return f["fmt"]

        return None




class SecretMonkee:  # --------------------------------------------------------------------
    def __init__(self, secretclient, project_id, what_for=[]):
        self.secretclient = secretclient
        self.project_id = project_id
        self.secrets = {}
        self.whatFor = what_for
        self.getSecrets()

    def _addSecret(self, secretName, secret):
        self.secrets[secretName] = secret

    def getSecret(self, secretclient, project_id, secret_name):
        request = {
            "name": f"projects/{project_id}/secrets/{secret_name}/versions/latest"}
        response = secretclient.access_secret_version(request)
        secret_string = response.payload.data.decode("UTF-8")
        return secret_string

    def getSecrets(self):
        """
        Gets the secrets for the given whatFors
            ['sql','neo','redis].

            USAGE:
            sm = SecretMonkee(secretclient, project_id, ['neo']).secrets
        """
        if 'sql' in self.whatFor:
            self._addSecret("db_user", self.getSecret(
                self.secretclient, self.project_id, "DB_USER"))
            self._addSecret("db_pass", self.getSecret(
                self.secretclient, self.project_id, "DB_PASS"))
            self._addSecret("db_name", self.getSecret(
                self.secretclient, self.project_id, "DB_NAME"))
            self._addSecret("cloud_sql_connection_name", self.getSecret(
                self.secretclient, self.project_id, "CLOUD_SQL_CONNECTION_NAME"
            ))

        if 'neo' in self.whatFor:
            self._addSecret("neo_uri", self.getSecret(
                self.secretclient, self.project_id, "DB_NEO_URI"))
            self._addSecret("neo_user", self.getSecret(
                self.secretclient, self.project_id, "DB_NEO_USER"))
            self._addSecret("neo_pass", self.getSecret(
                self.secretclient, self.project_id, "DB_NEO_PASS"))

        if 'redis' in self.whatFor:
            self._addSecret("redis_host", self.getSecret(
                self.secretclient, self.project_id, "REDISHOST"))
            self._addSecret("redis_port", self.getSecret(
                self.secretclient, self.project_id, "REDISPORT"))
        return self.secrets




import requests
from dateutil import parser
import json
from datetime import datetime, timezone
import time
import sys
import random
import uuid
import copy


# --------------------------------------------------------------------


class RoundTripEncoder(json.JSONEncoder):
    DATE_FORMAT = "%Y-%m-%d"
    TIME_FORMAT = "%H:%M:%S.%f"

    def default(self, obj):
        if isinstance(obj, datetime):
            return {
                "_type": "datetime",
                "value": obj.strftime("%s %s" % (self.DATE_FORMAT, self.TIME_FORMAT)),
            }
        return super(RoundTripEncoder, self).default(obj)


class RoundTripDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "_type" not in obj:
            return obj
        type = obj["_type"]
        if type == "datetime":
            return parser.parse(obj["value"])
        return obj


def call_cloud_function(url, headers, data):
    """
    Calls the cloud function at url with headers and data passed as part of the POST.
    Returns JSON response, passed through RoundTripDecoder
    """

    response_data = None

    try:
        response = requests.post(url=url, data=data, headers=headers)
        response_data = json.loads(response.text, cls=RoundTripDecoder)
    except Exception as e:
        print("ERROR in call_cloud_function: {}".format(str(e)))

    return response_data


class MonkeeTracker:  # --------------------------------------------------------------------
    def __init__(self, db, app_name, function_name, human_uid):
        self.tracker = []
        self.db = db
        self.app_name = app_name
        self.human_uid = human_uid
        self.function_name = function_name

    def set_db(self, db_destination):
        self.db = db_destination

    def set_human_uid(self, human_uid):
        self.human_uid = human_uid

    def track(self, event_name):
        self.tracker.append(
            {"ts": datetime.now(timezone.utc), "e": event_name})

    def persist(self, logUid=None):
        self.track("persist_tracker")
        if len(self.tracker) >= 1:
            diff = self.tracker[-1]["ts"] - self.tracker[0]["ts"]
        overall_diff_s = diff.seconds + diff.microseconds / 1000000

        ti = 0
        while ti < len(self.tracker) - 1:
            next_time = self.tracker[ti + 1]["ts"]
            elapsed_mcs = next_time - self.tracker[ti]["ts"]
            self.tracker[ti]["elapsed_time"] = (
                elapsed_mcs.seconds + elapsed_mcs.microseconds / 1000000
            )
            self.tracker[ti]["a_perc"] = (
                str(
                    int(
                        round(
                            100 *
                            self.tracker[ti]["elapsed_time"] /
                            overall_diff_s, 0
                        )
                    )
                )
                + "%"
            )
            ti += 1

        track_dict = {
            "humanUid": self.human_uid,
            "calledFrom": self.function_name,
            "duration": overall_diff_s,
            "log": self.tracker,
        }
        r = int(10000 * random.random())
        seconds = 9999999999 - time.time()
        logUid = str(seconds) + str(r)
        if logUid is None:
            self.db.document().set(track_dict)
        else:
            self.db.document(logUid).set(track_dict)


def get_size(json_obj):
    """
    returns the size of the JSON object in bytes
    """
    dumps = json.dumps(json_obj, cls=RoundTripEncoder)
    size_bytes = sys.getsizeof(dumps)
    return size_bytes


def dateDiff(unit, ts1, ts2):
    """
    returns the time delta between ts1 and ts2 in the provided unit.
    Unit in: ['second','minute','hour','day']
    """
    elapsedTime = ts2 - ts1
    totalSeconds = elapsedTime.total_seconds()

    if unit in ["s", "sec", "second"]:
        return totalSeconds
    elif unit in ["mn", "min", "minute"]:
        return totalSeconds / 60
    elif unit in ["hr", "hour"]:
        return totalSeconds / 60 / 60
    elif unit in ["d", "day"]:
        return totalSeconds / 60 / 60 / 24


def getval(dictionary, key, default_value=None):
    if dictionary is not None:
        if key in dictionary:
            ret = dictionary[key]
        else:
            ret = default_value
    else:
        ret = default_value
    return ret


def get_uuid():
    return str(uuid.uuid4())


def describe_time(hh_ago):
    ret = ""
    hh_ago = int(round(hh_ago))
    if hh_ago == 0:
        ret = "very recently"
    elif hh_ago == 1:
        ret = "an hour ago"
    elif hh_ago <= 24:
        ret = str(hh_ago) + " hours ago"
    elif hh_ago <= 48:
        ret = "yesterday"
    else:
        dd_ago = int(round(hh_ago / 24))
        ret = str(dd_ago) + " days ago"

    return ret


def makeAscendingUid():
    """
    Creates a uid such that new uids are always alphabetically in front of older ones.
    For typical use in creating FB doc UIDs such that new docs will show up at the top of the collection
    """
    docUid = str(1625607464 * 3 - int(time.time()))
    uuid = get_uuid()
    return f'{docUid}_{uuid}'
